/* diffpop v0.9 (March 2011)
 * Copyright (C) 2011 Norbert de Jonge <mail@norbertdejonge.nl>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see [ www.gnu.org/licenses/ ].
 *
 * To properly read this code, set your program's tab stop to: 2.
 */

/*========== Includes ==========*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
/*========== Includes ==========*/

/*========== Defines ==========*/
#define EXIT_ERROR 1
#define EXIT_NORMAL 0
#define MAX_DATA 720
#define MAX_ATTR 100
#define MAX_REMEMBER 500
#define MAX_VERSION 10
#define PROG_NAME "diffpop"
#define PROG_VERSION "v0.9 (March 2011)"
#define COPYRIGHT "Copyright (C) 2011 Norbert de Jonge"
#define RV_NAME 0
#define RV_TYPE 1
#define RV_DEFAULT 2
#define RV_SIZE 3
#define RV_READOP 4
#define ESCAPE 27
/*========== Defines ==========*/

int iDebug;
int iAttributes;
int iRememberedValues;
int iCheat;
int iDifferences;
int iContinueAfterCheckFailed;
int iLocations;
int iBold;
char sType[MAX_DATA + 2];
char sAttributes[2 + 2][MAX_ATTR + 2][MAX_DATA + 2];
char sRememberedValues[4 + 2][MAX_REMEMBER + 2][MAX_DATA + 2];
char sFolder[MAX_DATA + 2];
char sGroup[MAX_DATA + 2];
char sHack[MAX_DATA + 2];

int ReadFromFile (int iFd, char *sWhat, int iSize, unsigned char *sRetString);
void Parse (char *sXML);
char *GetValue (char *sName);
void Check (char *sHex, int iFdEXE);
void RememberValues (char *sValueName, char *sValueType, char *sValueDefault,
	char *sValueSize, char *sValueReadoperation);
void ReadAndDisplay (char *sValueName, char *sValueType, char *sValueDefault,
	char *sValueSize, char *sValueReadoperation, int iFdEXE);
void ShowLocation (void);
char *strtolower(char *str);
void ShowUsage (void);

/*****************************************************************************/
int main (int argc, char *argv[])
/*****************************************************************************/
{
	int iFdXML;
	int iFdEXE;
	unsigned char sXML[MAX_DATA + 2];
	char sVersion[MAX_VERSION + 2];
	unsigned long lOffset;
	char sValueName[MAX_DATA + 2];
	char sValueType[MAX_DATA + 2];
	char sValueDefault[MAX_DATA + 2];
	char sValueSize[MAX_DATA + 2];
	char sValueReadoperation[MAX_DATA + 2];
	int iTemp;
	int iOffset;
	int iReadingOffsets;
	struct stat stStat;
	int iXML;
	int iHaveExe;
	char sExeFile[MAX_DATA + 2];
	int iPacked;

	iDebug = 0;
	iLocations = 0;
	iBold = 0;
	iContinueAfterCheckFailed = 0;
	iHaveExe = 0;
	iCheat = 0;
	iDifferences = 0;
	iPacked = 0;
	snprintf (sVersion, MAX_VERSION, "%s", "");
	snprintf (sFolder, MAX_DATA, "%s", "");
	snprintf (sGroup, MAX_DATA, "%s", "");
	snprintf (sHack, MAX_DATA, "%s", "");
	iRememberedValues = -1;

	if (argc > 1)
	{
		for (iTemp = 1; iTemp <= argc - 1; iTemp++)
		{
			if ((strcmp (argv[iTemp], "-h") == 0) ||
				(strcmp (argv[iTemp], "-?") == 0) ||
				(strcmp (argv[iTemp], "--help") == 0))
			{
				ShowUsage();
			}
			else if ((strcmp (argv[iTemp], "-v") == 0) ||
				(strcmp (argv[iTemp], "--version") == 0))
			{
				printf ("%s %s\n", PROG_NAME, PROG_VERSION);
				exit (EXIT_NORMAL);
			}
			else if ((strcmp (argv[iTemp], "-d") == 0) ||
				(strcmp (argv[iTemp], "--debug") == 0))
			{
				iDebug = 1;
			}
			else if ((strcmp (argv[iTemp], "-c") == 0) ||
				(strcmp (argv[iTemp], "--continue") == 0))
			{
				iContinueAfterCheckFailed = 1;
			}
			else if ((strcmp (argv[iTemp], "-l") == 0) ||
				(strcmp (argv[iTemp], "--locations") == 0))
			{
				iLocations = 1;
			}
			else if ((strcmp (argv[iTemp], "-b") == 0) ||
				(strcmp (argv[iTemp], "--bold") == 0))
			{
				iBold = 1;
			}
			else if ((strcmp (argv[iTemp], "--exe=p0") == 0) ||
				(strcmp (argv[iTemp], "--exe=p3") == 0) ||
				(strcmp (argv[iTemp], "--exe=p4") == 0) ||
				(strcmp (argv[iTemp], "--exe=u0") == 0) ||
				(strcmp (argv[iTemp], "--exe=u3") == 0) ||
				(strcmp (argv[iTemp], "--exe=u4") == 0))
			{
				snprintf (sVersion, MAX_VERSION, "%c%c", argv[iTemp][6],
					argv[iTemp][7]);
			}
			else if (iHaveExe == 0)
			{
				snprintf (sExeFile, MAX_DATA, "%s", argv[iTemp]);
				iHaveExe = 1;
			}
			else
			{
				ShowUsage();
			}
		}
	}
	if ((argc == 1) || (iHaveExe == 0))
	{
		ShowUsage();
	}

	iFdXML = open ("xml/princehack.xml", O_RDONLY);
	if (iFdXML == -1)
	{
		printf ("[FAILED] Could not open princehack.xml: %s!\n",
			strerror (errno)); exit (EXIT_ERROR);
	}
	iFdEXE = open (sExeFile, O_RDONLY);
	if (iFdEXE == -1)
	{
		printf ("[FAILED] Could not open %s: %s!\n", sExeFile,
			strerror (errno)); exit (EXIT_ERROR);
	}

	/* Cheat, because princehack.xml contains "duplicated" before "name". */
	RememberValues ("level to load", "unsigned short int", "63558",
		"NONE", "if(63558,off,~);");
	RememberValues ("Quit when time expires", "unsigned char", "off",
		"NONE", "if(2,off,on);");

	if (strcmp (sVersion, "") == 0)
	{
		stat (sExeFile, &stStat);
		switch ((int)stStat.st_size)
		{
			case 123335:
				iPacked = 1;
				snprintf (sVersion, MAX_VERSION, "%s", "p0"); break;
			case 125115:
				iPacked = 1;
				snprintf (sVersion, MAX_VERSION, "%s", "p3"); break;
			case 110855:
				iPacked = 1;
				snprintf (sVersion, MAX_VERSION, "%s", "p4"); break;
			case 129504:
				snprintf (sVersion, MAX_VERSION, "%s", "u0"); break;
			case 129472:
				snprintf (sVersion, MAX_VERSION, "%s", "u3"); break;
			case 115008:
				snprintf (sVersion, MAX_VERSION, "%s", "u4"); break;
			default:
				printf ("[ INFO ] Strange file size. Assuming this is a p0"
					" executable.\n");
				printf ("[ INFO ] Start with -h (or -? or --help) for"
					" information about --exe=<TYPE>.\n");
				iPacked = 1;
				snprintf (sVersion, MAX_VERSION, "%s", "p0"); break;
		}
	}
	if (iPacked == 1)
	{
		printf ("[ INFO ] This is a packed executable. You may want to"
			" use UPackExe on it first.\n");
	}

	iReadingOffsets = 0;
	do {
		ReadFromFile (iFdXML, "", 0, sXML);
		Parse ((char *)sXML);
		iXML = strlen((char *)sXML);
		if (iDebug == 1)
			{ printf ("[ INFO ] Attributes: %i\n", iAttributes + 1); }
		if (strcmp (sType, "offset") == 0)
		{
			if (iReadingOffsets == 0) { iOffset = 0; }
			iReadingOffsets = 1;
			if (strcmp (GetValue ("file"), sVersion) == 0)
			{
				if (strcmp (GetValue ("value"), "none") == 0)
				{
					iOffset = 0;
				} else {
					lOffset = strtol (GetValue ("value"), NULL, 16);
					lseek (iFdEXE, lOffset, SEEK_SET);
					if (iDebug == 1)
						{ printf ("[ INFO ] Offset: %s\n", GetValue ("value")); }
					iOffset = 1;
				}
			}
		} else {
			iReadingOffsets = 0;
		}

		if (strcmp (sType, "folder") == 0)
			{ snprintf (sFolder, MAX_DATA, "%s", GetValue ("name")); }
		if (strcmp (sType, "/folder") == 0)
			{ snprintf (sFolder, MAX_DATA, "%s", ""); }
		if (strcmp (sType, "group") == 0)
			{ snprintf (sGroup, MAX_DATA, "%s", GetValue ("name")); }
		if (strcmp (sType, "/group") == 0)
			{ snprintf (sGroup, MAX_DATA, "%s", ""); }
		if (strcmp (sType, "hack") == 0)
			{ snprintf (sHack, MAX_DATA, "%s", GetValue ("name")); }
		if (strcmp (sType, "/hack") == 0)
			{ snprintf (sHack, MAX_DATA, "%s", ""); }

		if ((strcmp (sType, "check") == 0) && (iOffset == 1))
		{
			if (iDebug == 1)
				{ printf ("[ INFO ] Checking for: %s\n", GetValue ("codes")); }
			Check (GetValue ("codes"), iFdEXE);
		}
		if ((strcmp (sType, "read") == 0) && (iOffset == 1))
		{
			snprintf (sValueName, MAX_DATA, "%s", GetValue ("name"));
			snprintf (sValueType, MAX_DATA, "%s", GetValue ("type"));
			snprintf (sValueDefault, MAX_DATA, "%s", GetValue ("default"));
			snprintf (sValueSize, MAX_DATA, "%s", GetValue ("size"));
			snprintf (sValueReadoperation, MAX_DATA, "%s",
				GetValue ("readoperation"));
			if (strcmp (sValueName, "NONE") == 0)
			{
				if (iDebug == 1) { printf ("[ INFO ] No type found.\n"); }
				for (iTemp = 0; iTemp <= iRememberedValues; iTemp++)
				{
					if (strcmp (sRememberedValues[RV_NAME][iTemp],
						GetValue ("duplicated")) == 0)
					{
						snprintf (sValueName, MAX_DATA, "%s",
							sRememberedValues[RV_NAME][iTemp]);
						if (strcmp (sValueType, "NONE") == 0)
							snprintf (sValueType, MAX_DATA, "%s",
								sRememberedValues[RV_TYPE][iTemp]);
						if (strcmp (sValueDefault, "NONE") == 0)
							snprintf (sValueDefault, MAX_DATA, "%s",
								sRememberedValues[RV_DEFAULT][iTemp]);
						if (strcmp (sValueSize, "NONE") == 0)
							snprintf (sValueSize, MAX_DATA, "%s",
								sRememberedValues[RV_SIZE][iTemp]);
						if (strcmp (sValueReadoperation, "NONE") == 0)
							snprintf (sValueReadoperation, MAX_DATA, "%s",
								sRememberedValues[RV_READOP][iTemp]);
					}
				}
			}
			if (sXML[iXML - 2] == '/')
			{
				if ((strcmp (sValueName, "when level starts") == 0) ||
					(strcmp (sValueName, "WDA in palace") == 0) ||
					(strcmp (sValueName, "level to load") == 0) ||
					(strcmp (sValueName, "Quit when time expires") == 0))
					{ iCheat++; } else { iCheat = 0; }
				RememberValues (sValueName, sValueType, sValueDefault, sValueSize,
					sValueReadoperation);
				ReadAndDisplay (sValueName, sValueType, sValueDefault, sValueSize,
					sValueReadoperation, iFdEXE);
			}
		}
		if ((strcmp (sType, "option") == 0) && (iOffset == 1))
		{
			if (strcmp (GetValue ("default"), "true") == 0)
			{
				if (strcmp (GetValue("value"), "-1") == 0)
				{
					if (strcmp (sValueType, "signed short int") == 0)
					{
						/* Two bytes: 65536 - 1. */
						snprintf (sValueDefault, MAX_DATA, "%s", "65535");
					} else {
						/* One byte: 256 - 1. */
						snprintf (sValueDefault, MAX_DATA, "%s", "255");
					}
				} else {
					snprintf (sValueDefault, MAX_DATA, "%s", GetValue("value"));
				}
			}
		}
		if ((strcmp (sType, "/read") == 0) && (iOffset == 1))
		{
			if ((strcmp (sValueName, "when level starts") == 0) ||
				(strcmp (sValueName, "WDA in palace") == 0) ||
				(strcmp (sValueName, "level to load") == 0) ||
				(strcmp (sValueName, "Quit when time expires") == 0))
				{ iCheat++; } else { iCheat = 0; }
			RememberValues (sValueName, sValueType, sValueDefault, sValueSize,
				sValueReadoperation);
			ReadAndDisplay (sValueName, sValueType, sValueDefault, sValueSize,
				sValueReadoperation, iFdEXE);
		}
	} while (strcmp (sType, "/princehack") != 0);

	close (iFdXML);
	close (iFdEXE);

	printf ("[ INFO ] Found %i differences.\n", iDifferences);

	return 0;
}

/*****************************************************************************/
int ReadFromFile (int iFd, char *sWhat, int iSize, unsigned char *sRetString)
/*****************************************************************************/
{
	int iLength;
	int iRead;
	unsigned char sRead[1 + 2];
	int iEnd;
	int iSign;

	if (strlen (sWhat) != 0)
	{
		if (iDebug == 1) { printf ("[ INFO ] Loading: %s\n", sWhat); }
	}
	iLength = 0;
	iEnd = 0;
	iSign = -1;
	do {
		iRead = read (iFd, sRead, 1);
		switch (iRead)
		{
			case -1: printf ("[FAILED] Could not read: %s!\n", strerror (errno));
				exit (EXIT_ERROR); break;
			case 0: printf ("[ INFO ] End of file\n"); iEnd = 1; break;
			default:
				switch (iSize)
				{
					case 0:
						if (sRead[0] == '<')
						{
							if (iSign == -1) { iSign = 1; } else { iSign++; }
						}
						if (sRead[0] == '>')
						{
							iSign--;
							if (iSign == 0) { iEnd = 1; }
						}
						break;
					default:
						if (iLength + 1 == iSize) { iEnd = 1; }
						break;
				}
				if ((iSign != -1) || (iSize != 0))
				{
					sRetString[iLength] = sRead[0];
					iLength++;
				}
				break;
		}
	} while (iEnd == 0);
	sRetString[iLength] = '\0';

	return (iLength);
}

/*****************************************************************************/
void Parse (char *sXML)
/*****************************************************************************/
{
	int iTemp;
	int iState;
	char cChar;
	char sString[MAX_DATA + 2];
	char sTemp[MAX_DATA + 2];

	if (iDebug == 1) { printf ("[ INFO ] XML: %s\n", sXML); }
	if (sXML[0] != '<')
	{
		printf ("[FAILED] Incorrect XML data!\n");
		exit (EXIT_ERROR);
	} else {
		if (sXML[1] == '!')
		{
			snprintf (sType, MAX_DATA, "%s", "comment");
		} else {
			iState = 0;
			iAttributes = -1;
			snprintf (sString, MAX_DATA, "%s", "");
			for (iTemp = 1; iTemp < (int)strlen (sXML); iTemp++)
			{
				cChar = sXML[iTemp];
				switch (iState)
				{
					case 0: /* type */
						if ((cChar == ' ') || (cChar == '>'))
						{
							snprintf (sType, MAX_DATA, "%s", sString);
							snprintf (sString, MAX_DATA, "%s", "");
							iState = 1;
						} else {
							snprintf (sTemp, MAX_DATA, "%s", sString);
							snprintf (sString, MAX_DATA, "%s%c", sTemp, cChar);
						}
						break;
					case 1: /* name */
					case 2:
						switch (cChar)
						{
							case ' ': break;
							case '=':
								iState = 2;
								iAttributes++;
								break;
							case '"':
								snprintf (sAttributes[0][iAttributes], MAX_DATA, "%s",
									sString);
								if (iDebug == 1) { printf ("[ INFO ] Name (%i): %s\n",
									iAttributes + 1, sAttributes[0][iAttributes]); }
								snprintf (sString, MAX_DATA, "%s", "");
								iState = 3;
								break;
							default:
								snprintf (sTemp, MAX_DATA, "%s", sString);
								snprintf (sString, MAX_DATA, "%s%c", sTemp, cChar);
								break;
						}
						break;
					case 3: /* value */
						switch (cChar)
						{
							case '"':
								snprintf (sAttributes[1][iAttributes], MAX_DATA, "%s",
									sString);
								if (iDebug == 1) { printf ("[ INFO ] Value (%i): %s\n",
									iAttributes + 1, sAttributes[1][iAttributes]); }
								snprintf (sString, MAX_DATA, "%s", "");
								iState = 1;
								break;
							default:
								snprintf (sTemp, MAX_DATA, "%s", sString);
								snprintf (sString, MAX_DATA, "%s%c", sTemp, cChar);
								break;
						}
						break;
				}
			}
		}
	}
}

/*****************************************************************************/
char *GetValue (char *sName)
/*****************************************************************************/
{
	int iTemp;

	for (iTemp = 0; iTemp <= iAttributes; iTemp++)
	{
		if (strcmp (sAttributes[0][iTemp], sName) == 0)
		{
			return (sAttributes[1][iTemp]);
		}
	}
	return ("NONE");
}

/*****************************************************************************/
void Check (char *sHex, int iFdEXE)
/*****************************************************************************/
{
	int iTemp;
	char sString[MAX_DATA + 2];
	unsigned char sOne[1 + 2];
	unsigned long lXML;
	unsigned long lEXE;

	iTemp = 0;
	do {
		snprintf (sString, MAX_DATA, "0x%c%c", sHex[iTemp], sHex[iTemp + 1]);
		lXML = strtol (sString, NULL, 16);
		ReadFromFile (iFdEXE, "", 1, sOne);
		snprintf (sString, MAX_DATA, "%02x", sOne[0]);
		lEXE = strtol (sString, NULL, 16);
		if ((sHex[iTemp] != '.') && (lXML != lEXE))
		{
			printf ("[FAILED] Check failed: %lu is not %lu!\n", lXML, lEXE);
			if (iContinueAfterCheckFailed != 1)
			{
				printf ("[ INFO ] Start with -c (or --continue) to ignore this.\n");
				exit (EXIT_ERROR);
			}
		}
		iTemp = iTemp + 3;
	} while (iTemp < (int)strlen (sHex));
}

/*****************************************************************************/
void RememberValues (char *sValueName, char *sValueType, char *sValueDefault,
	char *sValueSize, char *sValueReadoperation)
/*****************************************************************************/
{
	iRememberedValues++;
	snprintf (sRememberedValues[RV_NAME][iRememberedValues],
		MAX_DATA, "%s", sValueName);
	snprintf (sRememberedValues[RV_TYPE][iRememberedValues],
		MAX_DATA, "%s", sValueType);
	snprintf (sRememberedValues[RV_DEFAULT][iRememberedValues],
		MAX_DATA, "%s", sValueDefault);
	snprintf (sRememberedValues[RV_SIZE][iRememberedValues],
		MAX_DATA, "%s", sValueSize);
	snprintf (sRememberedValues[RV_READOP][iRememberedValues],
		MAX_DATA, "%s", sValueReadoperation);
}

/*****************************************************************************/
void ReadAndDisplay (char *sValueName, char *sValueType, char *sValueDefault,
	char *sValueSize, char *sValueReadoperation, int iFdEXE)
/*****************************************************************************/
{
	int iBytes;
	char sString[MAX_DATA + 2];
	char sTemp[MAX_DATA + 2];
	unsigned long luNumber;
	unsigned char sData[MAX_DATA + 2];
	char sValueReal[MAX_DATA + 2];
	int iTemp;
	int iDefault;

	if (strcmp (sValueType, "unsigned short int") == 0) { iBytes = 2; }
	if (strcmp (sValueType, "unsigned char") == 0) { iBytes = 1; }
	if (strcmp (sValueType, "array") == 0) { iBytes = 24; }
	if (strcmp (sValueType, "char") == 0) { iBytes = 1; }
	if (strcmp (sValueType, "signed short int") == 0) { iBytes = 2; }
	if (strcmp (sValueType, "string") == 0)
	{
		iBytes = atoi(sValueSize);
		if (iBytes == 0) { printf ("[FAILED] Size is 0!\n"); exit (EXIT_ERROR); }
	}
	if (strcmp (sValueType, "short") == 0) { iBytes = 2; }
	if (strcmp (sValueType, "signed char") == 0) { iBytes = 1; }
	if (strcmp (sValueType, "boolean") == 0) { iBytes = 1; }
	if (strcmp (sValueType, "long") == 0) { iBytes = 4; }

	ReadFromFile (iFdEXE, "", iBytes, sData);

	if (strcmp (sValueType, "string") != 0)
	{
		snprintf (sString, MAX_DATA, "%s", "");
		for (iTemp = iBytes - 1; iTemp >= 0; iTemp--)
		{
			snprintf (sTemp, MAX_DATA, "%s%02x", sString, sData[iTemp]);
			snprintf (sString, MAX_DATA, "%s", sTemp);
		}
	}

	if (strcmp (sValueType, "string") == 0)
	{
		if (strcmp (sValueReadoperation, "tolower;") == 0)
		{
			snprintf (sValueReal, MAX_DATA, "%s", strtolower ((char *)sData));
		} else {
			snprintf (sValueReal, MAX_DATA, "%s", (char *)sData);
		}
	} else {
		luNumber = strtoul (sString, NULL, 16);
/*
		if (strcmp (sValueReadoperation, "add(1);") == 0)
			{ luNumber++; }
		if (strcmp (sValueReadoperation, "subtract(1);") == 0)
			{ luNumber--; }
*/
		snprintf (sValueReal, MAX_DATA, "%lu", luNumber);
	}

	if (strcmp (sValueName, "when level starts") == 0)
	{
		if (strcmp (sValueReal, "4294967295") == 0)
			snprintf (sValueReal, MAX_DATA, "%s", "1");
				else snprintf (sValueReal, MAX_DATA, "%s", "0");
	}
	if (strcmp (sValueName, "WDA in palace") == 0)
	{
		iDefault = 0;
		switch (iCheat)
		{
			/* Cheat, because princehack.xml lacks proper readoperations. */
			case 1: if (strcmp (sValueReal, "116") == 0) { iDefault = 1; } break;
			case 2: if (strcmp (sValueReal, "27") == 0) { iDefault = 1; } break;
			case 3: if (strcmp (sValueReal, "192") == 0) { iDefault = 1; } break;
			case 4: if (strcmp (sValueReal, "247") == 0) { iDefault = 1; } break;
			case 5: if (strcmp (sValueReal, "216") == 0) { iDefault = 1; } break;
			case 6: if (strcmp (sValueReal, "116") == 0) { iDefault = 1; } break;
		}
		if (iDefault == 1) { snprintf (sValueReal, MAX_DATA, "%s", "1"); }
			else { snprintf (sValueReal, MAX_DATA, "%s", "0"); }
	}
	if (strcmp (sValueName, "level to load") == 0)
	{
		iDefault = 0;
		switch (iCheat)
		{
			/* Cheat, because princehack.xml lacks proper readoperations. */
			case 1: if (strcmp (sValueReal, "54") == 0) { iDefault = 1; } break;
			case 2: if (strcmp (sValueReal, "38") == 0) { iDefault = 1; } break;
			case 3: if (strcmp (sValueReal, "139") == 0) { iDefault = 1; } break;
			case 4: if (strcmp (sValueReal, "63558") == 0) { iDefault = 1; } break;
		}
		if (iDefault == 1) { snprintf (sValueReal, MAX_DATA, "%s", "63558"); }
			else { snprintf (sValueReal, MAX_DATA, "%s", "0"); }
	}
	if (strcmp (sValueName, "Quit when time expires") == 0)
	{
		iDefault = 0;
		switch (iCheat)
		{
			/* Cheat, because princehack.xml lacks proper readoperations. */
			case 1: if (strcmp (sValueReal, "148") == 0) { iDefault = 1; } break;
			case 2: if (strcmp (sValueReal, "79") == 0) { iDefault = 1; } break;
			case 3: if (strcmp (sValueReal, "2") == 0) { iDefault = 1; } break;
		}
		if (iDefault == 1) { snprintf (sValueReal, MAX_DATA, "%s", "off"); }
			else { snprintf (sValueReal, MAX_DATA, "%s", "0"); }
	}

	if (strcmp (sValueReal, sValueDefault) != 0)
	{
		if (iLocations == 1) { ShowLocation(); }
		if (iBold == 1) { printf ("%c[1m", ESCAPE); }
		printf ("[ DIFF ] %s: %s (default: %s", sValueName, sValueReal,
			sValueDefault);
		if (strcmp (sValueReadoperation, "NONE") != 0)
		{
			printf ("; readoperation: %s", sValueReadoperation);
		}
		printf (")\n");
		if (iBold == 1) { printf ("%c[0m", ESCAPE); }
		iDifferences++;
	} else {
		if (iDebug == 1)
		{
			if (iLocations == 1) { ShowLocation(); }
			printf ("[ SAME ] %s: %s\n", sValueName, sValueReal);
		}
	}
}

/*****************************************************************************/
void ShowLocation (void)
/*****************************************************************************/
{
	printf ("[ INFO ] Location: %s", sFolder);
	if (strcmp (sGroup, "") != 0) { printf (" (%s)", sGroup); }
	if (strcmp (sHack, "") != 0) { printf (" (%s)", sHack); }
	printf ("\n");
}

/*****************************************************************************/
char *strtolower (char *sString)
/*****************************************************************************/
{
	int iTemp;

	for (iTemp = 0; iTemp < (int)strlen (sString); iTemp++)
	{
		if (sString[iTemp] >= 'A' && sString[iTemp] <= 'Z')
			sString[iTemp] = sString[iTemp] + 32;
	}
	return (sString);
}

/*****************************************************************************/
void ShowUsage (void)
/*****************************************************************************/
{
	printf ("%s %s\n%s\n\n", PROG_NAME, PROG_VERSION, COPYRIGHT);
	printf ("Usage:\n");
	printf ("  diffpop [OPTIONS] <PRINCE.EXE>\n\nOptions:\n");
	printf ("  -h, -?,    --help           display this help and exit\n");
	printf ("  -v,        --version        output version information and"
		" exit\n");
	printf ("  -d,        --debug          show debug information\n");
	printf ("  -c,        --continue       continue even after failed"
		" checks\n");
	printf ("  -l,        --locations      show additional location"
		" information\n");
	printf ("  -b,        --bold           print the differences in a bold"
		" font\n");
	printf ("             --exe=<TYPE>     specify the executable's type"
		" (default: autodetect)\n");
	printf ("\nTypes:\n");
	printf ("  p0, p3, p4, u0, u3, u4\n");
	exit (EXIT_NORMAL);
}
